@@ -103,7 +103,7 @@ void destroy_ecpp_gcds(void) {
 
 
 
-static int check_for_factor2(mpz_t f, mpz_t inputn, mpz_t fmin, mpz_t n, int stage, mpz_t* sfacs, int* nsfacs, int degree)
+static int check_for_factor(mpz_t f, mpz_t inputn, mpz_t fmin, mpz_t n, int stage, mpz_t* sfacs, int* nsfacs, int degree)
 {
   int success, sfaci;
   UV B1;
@@ -113,20 +113,6 @@ static int check_for_factor2(mpz_t f, mpz_t inputn, mpz_t fmin, mpz_t n, int sta
 
   if (mpz_cmp(n, fmin) <= 0) return 0;
 
-#if 0
-  {
-    /* Straightforward trial division up to 3000. */
-    PRIME_ITERATOR(iter);
-    UV tf;
-    UV const trial_limit = 3000;
-    for (tf = 2; tf < trial_limit; tf = prime_iterator_next(&iter)) {
-      if (mpz_cmp_ui(n, tf*tf) < 0) break;
-      while (mpz_divisible_ui_p(n, tf))
-        mpz_divexact_ui(n, n, tf);
-    }
-    prime_iterator_destroy(&iter);
-  }
-#else
   /* Utilize GMP's fast gcd algorithms.  Trial to 224737 with two gcds. */
   mpz_tdiv_q_2exp(n, n, mpz_scan1(n, 0));
   while (mpz_divisible_ui_p(n, 3))  mpz_divexact_ui(n, n, 3);
@@ -143,20 +129,6 @@ static int check_for_factor2(mpz_t f, mpz_t inputn, mpz_t fmin, mpz_t n, int sta
     mpz_divexact(n, n, f);
     mpz_gcd(f, n, _gcd_large);
   }
-  /* Quick stage 1 n-1 using a single big powm + gcd. */
-  if (stage == 0) {
-    if (mpz_cmp(n, fmin) <= 0) return 0;
-    mpz_set_ui(f, 2);
-    mpz_powm(f, f, _lcm_small, n);
-    mpz_sub_ui(f, f, 1);
-    mpz_gcd(f, f, n);
-    if (mpz_cmp_ui(f, 1) != 0 && mpz_cmp(f, n) != 0) {
-      mpz_divexact(n, n, f);
-      if (mpz_cmp(f, n) > 0)
-        mpz_set(n, f);
-    }
-  }
-#endif
 
   sfaci = 0;
   success = 1;
@@ -168,8 +140,8 @@ static int check_for_factor2(mpz_t f, mpz_t inputn, mpz_t fmin, mpz_t n, int sta
 
     success = 0;
     B1 = 300 + 3 * nsize;
-    if (degree <= 2) B1 += nsize; /* D1 & D2 are cheap to prove.  Encourage. */
-    if (degree <= 0) B1 += nsize; /* N-1 and N+1 are really cheap. */
+    if (degree <= 2) B1 += nsize;             /* D1 & D2 are cheap to prove. */
+    if (degree <= 0) B1 += 2*nsize;         /* N-1 and N+1 are really cheap. */
     if (degree > 20 && stage <= 1) B1 -= nsize;   /* Less time on big polys. */
     if (degree > 40) B1 -= nsize/2;               /* Less time on big polys. */
     if (stage >= 1) {
@@ -206,19 +178,19 @@ static int check_for_factor2(mpz_t f, mpz_t inputn, mpz_t fmin, mpz_t n, int sta
         if (!success) success = _GMP_pminus1_factor(n, f, 6*B1, 60*B1);
         /* p+1 with different initial point and searching farther */
         if (!success) success = _GMP_pplus1_factor(n, f, 1, B1/2, B1/2);
-        if (!success) success = _GMP_ecm_factor_projective(n, f, 250, 3);
+        if (!success) success = _GMP_ecm_factor_projective(n, f, 250, 2500, 3);
       } else if (stage == 3) {
         if (!success) success = _GMP_pbrent_factor(n, f, nsize+1, 16384);
         if (!success) success = _GMP_pminus1_factor(n, f, 60*B1, 600*B1);
         /* p+1 with a third initial point and searching farther */
         if (!success) success = _GMP_pplus1_factor(n, f, 2, 1*B1, 1*B1);
-        if (!success) success = _GMP_ecm_factor_projective(n, f, B1/4, 4);
+        if (!success) success = _GMP_ecm_factor_projective(n, f, B1/4, B1*4, 4);
       } else if (stage == 4) {
         if (!success) success = _GMP_pminus1_factor(n, f, 300*B1, 300*20*B1);
-        if (!success) success = _GMP_ecm_factor_projective(n, f, B1/2, 4);
+        if (!success) success = _GMP_ecm_factor_projective(n, f, B1/2, B1*8, 4);
       } else if (stage >= 5) {
         UV B = B1 * (stage-4) * (stage-4) * (stage-4);
-        if (!success) success = _GMP_ecm_factor_projective(n, f, B, 8+stage);
+        if (!success) success = _GMP_ecm_factor_projective(n, f, B, 10*B, 8+stage);
       }
     }
     if (success) {
@@ -243,81 +215,6 @@ static int check_for_factor2(mpz_t f, mpz_t inputn, mpz_t fmin, mpz_t n, int sta
   return -1;
 }
 
-/* return false or true with f (prime) > fmin */
-/* Set up for clever FPS, where the reduced n is remembered each time */
-static int check_for_factor(mpz_t f, mpz_t inputn, mpz_t fmin, mpz_t n, long stage)
-{
-  int success;
-
-  /* Use this so we don't modify their input value */
-  mpz_set(n, inputn);
-
-  if (mpz_cmp(n, fmin) <= 0) return 0;
-
-  if (stage == 1) {
-    /* simple trial division */
-    while (mpz_divisible_ui_p(n,  2)) mpz_divexact_ui(n, n,  2);
-    while (mpz_divisible_ui_p(n,  3)) mpz_divexact_ui(n, n,  3);
-    if (mpz_gcd_ui(NULL, n, 2850092245UL) != 1) {
-      while (mpz_divisible_ui_p(n,  5)) mpz_divexact_ui(n, n,  5);
-      while (mpz_divisible_ui_p(n,  7)) mpz_divexact_ui(n, n,  7);
-      while (mpz_divisible_ui_p(n, 11)) mpz_divexact_ui(n, n, 11);
-      while (mpz_divisible_ui_p(n, 13)) mpz_divexact_ui(n, n, 13);
-      while (mpz_divisible_ui_p(n, 17)) mpz_divexact_ui(n, n, 17);
-      while (mpz_divisible_ui_p(n, 19)) mpz_divexact_ui(n, n, 19);
-      while (mpz_divisible_ui_p(n, 41)) mpz_divexact_ui(n, n, 41);
-      while (mpz_divisible_ui_p(n, 43)) mpz_divexact_ui(n, n, 43);
-    }
-    if (mpz_gcd_ui(NULL, n, 2392308223UL) != 1) {
-      while (mpz_divisible_ui_p(n, 23)) mpz_divexact_ui(n, n, 23);
-      while (mpz_divisible_ui_p(n, 29)) mpz_divexact_ui(n, n, 29);
-      while (mpz_divisible_ui_p(n, 31)) mpz_divexact_ui(n, n, 31);
-      while (mpz_divisible_ui_p(n, 37)) mpz_divexact_ui(n, n, 37);
-      while (mpz_divisible_ui_p(n, 53)) mpz_divexact_ui(n, n, 53);
-      while (mpz_divisible_ui_p(n, 59)) mpz_divexact_ui(n, n, 59);
-    }
-  }
-
-  success = 1;
-  while (success) {
-
-    if (mpz_cmp(n, fmin) <= 0) return 0;
-    if (_GMP_is_prob_prime(n)) { mpz_set(f, n); return (mpz_cmp(f, fmin) > 0); }
-
-    success = 0;
-    if (stage == 1) {
-      /* We should probably do something like B1 = 10 * ndigits */
-      UV B1 = (mpz_sizeinbase(n, 2) > 1200) ? 6000 : 3000;
-      if (!success) success = _GMP_pminus1_factor(n, f, B1, 10*B1);
-    } else if (stage == 2) {
-      if (!success) success = _GMP_pminus1_factor(n, f, 10000, 200000);
-      if (!success) success = _GMP_ecm_factor_projective(n, f, 250, 4);
-    } else if (stage == 3) {
-      if (!success) success = _GMP_pminus1_factor(n, f, 40000, 800000);
-      if (!success) success = _GMP_ecm_factor_projective(n, f, 500, 4);
-    } else if (stage == 4) {
-      if (!success) success = _GMP_pminus1_factor(n, f, 100000, 5000000);
-      if (!success) success = _GMP_ecm_factor_projective(n, f, 1000, 10);
-    } else if (stage >= 5) {
-      UV B = 4000 * (stage-4) * (stage-4) * (stage-4);
-      if (!success) success = _GMP_ecm_factor_projective(n, f, B, 20);
-    }
-    if (success) {
-      if (mpz_cmp_ui(f, 1) == 0 || mpz_cmp(f, n) == 0) {
-        gmp_printf("factoring %Zd resulted in factor %Zd\n", n, f);
-        croak("internal error in ECPP factoring");
-      }
-      /* Is the factor f what we want? */
-      if ( mpz_cmp(f, fmin) > 0 && _GMP_is_prob_prime(f) )  return 1;
-      /* Divide out f */
-      mpz_divexact(n, n, f);
-    }
-  }
-  /* n is larger than fmin and not prime */
-  mpz_set(f, n);
-  return -1;
-}
-
 /* See:
  *   (1) Kaltofen, Valente, Yui 1989
  *   (2) Valente 1992 (Thesis)
@@ -595,13 +492,13 @@ static int find_curve(mpz_t a, mpz_t b, mpz_t x, mpz_t y,
 
 /* Select the 2, 4, or 6 numbers we will try to factor. */
 static void choose_m(mpz_t* mlist, long D, mpz_t u, mpz_t v, mpz_t N,
-                     mpz_t t, mpz_t Nminus1)
+                     mpz_t t, mpz_t Nplus1)
 {
-  int i;
-  mpz_add_ui(Nminus1, N, 1);
+  int i, j;
+  mpz_add_ui(Nplus1, N, 1);
 
-  mpz_add(mlist[0], Nminus1, u);
-  mpz_sub(mlist[1], Nminus1, u);
+  mpz_sub(mlist[0], Nplus1, u);     /* N+1-u */
+  mpz_add(mlist[1], Nplus1, u);     /* N+1+u */
   for (i = 2; i < 6; i++)
     mpz_set_ui(mlist[i], 0);
 
@@ -610,22 +507,28 @@ static void choose_m(mpz_t* mlist, long D, mpz_t u, mpz_t v, mpz_t N,
     mpz_mul_si(t, v, 3);
     mpz_add(t, t, u);
     mpz_tdiv_q_2exp(t, t, 1);
-    mpz_add(mlist[2], Nminus1, t);
-    mpz_sub(mlist[3], Nminus1, t);
+    mpz_sub(mlist[2], Nplus1, t);   /* N+1-(u+3v)/2 */
+    mpz_add(mlist[3], Nplus1, t);   /* N+1+(u+3v)/2 */
     mpz_mul_si(t, v, -3);
     mpz_add(t, t, u);
     mpz_tdiv_q_2exp(t, t, 1);
-    mpz_add(mlist[4], Nminus1, t);
-    mpz_sub(mlist[5], Nminus1, t);
+    mpz_sub(mlist[4], Nplus1, t);   /* N+1-(u-3v)/2 */
+    mpz_add(mlist[5], Nplus1, t);   /* N+1+(u-3v)/2 */
   } else if (D == -4) {
     mpz_mul_ui(t, v, 2);
-    mpz_add(mlist[2], Nminus1, t);
-    mpz_sub(mlist[3], Nminus1, t);
+    mpz_sub(mlist[2], Nplus1, t);   /* N+1-2v */
+    mpz_add(mlist[3], Nplus1, t);   /* N+1+2v */
   }
   /* m must not be prime */
   for (i = 0; i < 6; i++)
     if (mpz_sgn(mlist[i]) && _GMP_is_prob_prime(mlist[i]))
       mpz_set_ui(mlist[i], 0);
+  /* Sort the m values so we test the smallest first */
+  for (i = 0; i < 5; i++)
+    if (mpz_sgn(mlist[i]))
+      for (j = i+1; j < 6; j++)
+        if (mpz_sgn(mlist[j]) && mpz_cmp(mlist[i],mlist[j]) > 0)
+          mpz_swap( mlist[i], mlist[j] );
 }
 
 
@@ -641,8 +544,15 @@ static void choose_m(mpz_t* mlist, long D, mpz_t u, mpz_t v, mpz_t N,
  * found after stage 1, and we just try each of them.
  */
 
+#define VERBOSE_PRINT_N(step, ndigits, maxH, factorstage) \
+  if (verbose) { \
+    printf("%*sN[%d] (%d dig)", i, "", step, ndigits); \
+    if (factorstage > 1) printf(" [FS %d]", factorstage); \
+    fflush(stdout); \
+  }
+
 /* Recursive routine to prove via ECPP */
-static int ecpp_down(int i, mpz_t Ni, int facstage, IV* dlist, mpz_t* sfacs, int* nsfacs, char** prooftextptr)
+static int ecpp_down(int i, mpz_t Ni, int facstage, int *pmaxH, IV* dlist, mpz_t* sfacs, int* nsfacs, char** prooftextptr)
 {
   mpz_t a, b, u, v, m, q, minfactor, sqrtn, mD, t, t2;
   mpz_t mlist[6];
@@ -654,21 +564,20 @@ static int ecpp_down(int i, mpz_t Ni, int facstage, IV* dlist, mpz_t* sfacs, int
 
   nidigits = mpz_sizeinbase(Ni, 10);
 
-  k = _GMP_is_prob_prime(Ni);
-  if (k == 0)  return 0;
-  if (k == 2) {
+  downresult = _GMP_is_prob_prime(Ni);
+  if (downresult == 0)  return 0;
+  if (downresult == 2) {
     /* No need to put anything in the proof */
     if (verbose) printf("%*sN[%d] (%d dig)  PRIME\n", i, "", i, nidigits);
     return 2;
   }
-  downresult = 0;
-
-  if (verbose) {
-    printf("%*sN[%d] (%d dig)", i, "", i, nidigits);
-    if (facstage > 1) printf(" FS %d", facstage);
-    fflush(stdout);
+  if (i == 0 && facstage == 2 && _GMP_miller_rabin_random(Ni, 2, 0) == 0) {
+    gmp_printf("\n\n**** BPSW counter-example found?  ****\n**** N = %Zd ****\n\n", Ni);
+    return 0;
   }
 
+  VERBOSE_PRINT_N(i, nidigits, *pmaxH, facstage);
+
   mpz_init(a);  mpz_init(b);
   mpz_init(u);  mpz_init(v);
   mpz_init(m);  mpz_init(q);
@@ -698,20 +607,15 @@ static int ecpp_down(int i, mpz_t Ni, int facstage, IV* dlist, mpz_t* sfacs, int
         mpz_sub_ui(t2, sqrtn, 1);
         mpz_tdiv_q_2exp(t2, t2, 1);    /* t2 = minfactor */
 
-        facresult = check_for_factor2(q, m, t2, t, stage, sfacs, nsfacs, 0);
+        facresult = check_for_factor(q, m, t2, t, stage, sfacs, nsfacs, 0);
         if (facresult <= 0)
           continue;
         if (verbose)
           { printf(" n-1\n"); fflush(stdout); }
-        downresult = ecpp_down(i+1, q, next_stage, dlist, sfacs, nsfacs, prooftextptr);
-        if (downresult == 0)     /* composite */
-          goto end_down;
+        downresult = ecpp_down(i+1, q, next_stage, pmaxH, dlist, sfacs, nsfacs, prooftextptr);
+        if (downresult == 0) goto end_down;   /* composite */
         if (downresult == 1) {   /* nothing found at this stage */
-          if (verbose) {
-            printf("%*sN[%d] (%d dig)", i, "", i, nidigits);
-            if (facstage > 1) printf(" FS %d", facstage);
-            fflush(stdout);
-          }
+          VERBOSE_PRINT_N(i, nidigits, *pmaxH, facstage);
           continue;
         }
         if (verbose)
@@ -720,9 +624,8 @@ static int ecpp_down(int i, mpz_t Ni, int facstage, IV* dlist, mpz_t* sfacs, int
         if (verbose) { printf("  %d\n", curveresult); fflush(stdout); }
         if ( ! curveresult ) {
           /* This ought not happen */
-          dlist[dnum] = -2; /* skip this D value from now on */
-          if (verbose) gmp_printf("\n  Could not prove n-1 with N = %Zd\n", D, Ni);
-          downresult = 0;
+          if (verbose) gmp_printf("\n  Could not prove n-1 with N = %Zd\n", Ni);
+          downresult = 1;
           continue;
         }
         goto end_down;
@@ -732,20 +635,15 @@ static int ecpp_down(int i, mpz_t Ni, int facstage, IV* dlist, mpz_t* sfacs, int
         mpz_add_ui(t2, sqrtn, 1);
         mpz_tdiv_q_2exp(t2, t2, 1);    /* t2 = minfactor */
 
-        facresult = check_for_factor2(q, m, t2, t, stage, sfacs, nsfacs, 0);
+        facresult = check_for_factor(q, m, t2, t, stage, sfacs, nsfacs, 0);
         if (facresult <= 0)
           continue;
         if (verbose)
           { printf(" n+1\n"); fflush(stdout); }
-        downresult = ecpp_down(i+1, q, next_stage, dlist, sfacs, nsfacs, prooftextptr);
-        if (downresult == 0)     /* composite */
-          goto end_down;
+        downresult = ecpp_down(i+1, q, next_stage, pmaxH, dlist, sfacs, nsfacs, prooftextptr);
+        if (downresult == 0) goto end_down;   /* composite */
         if (downresult == 1) {   /* nothing found at this stage */
-          if (verbose) {
-            printf("%*sN[%d] (%d dig)", i, "", i, nidigits);
-            if (facstage > 1) printf(" FS %d", facstage);
-            fflush(stdout);
-          }
+          VERBOSE_PRINT_N(i, nidigits, *pmaxH, facstage);
           continue;
         }
         if (verbose)
@@ -754,9 +652,8 @@ static int ecpp_down(int i, mpz_t Ni, int facstage, IV* dlist, mpz_t* sfacs, int
         if (verbose) { printf("  %d\n", curveresult); fflush(stdout); }
         if ( ! curveresult ) {
           /* This ought not happen */
-          dlist[dnum] = -2; /* skip this D value from now on */
-          if (verbose) gmp_printf("\n  Could not prove n-1 with N = %Zd\n", D, Ni);
-          downresult = 0;
+          if (verbose) gmp_printf("\n  Could not prove n-1 with N = %Zd\n", Ni);
+          downresult = 1;
           continue;
         }
         goto end_down;
@@ -768,16 +665,8 @@ static int ecpp_down(int i, mpz_t Ni, int facstage, IV* dlist, mpz_t* sfacs, int
       /* Make sure we can get a class polynomial for this D. */
       poly_degree = poly_class_poly(D, NULL, &poly_type);
       if (poly_degree == 0)  continue;
-      /* We'll save time in the long run by not looking at big polys once
-       * we've found a good path from the start.  TODO: Needs more tuning. */
-      if (stage == 0 && i >= 0 && poly_degree > 2) break;
-      if (facstage == 1) {
-        if (i >  2 && nidigits < 1100 && poly_degree > 24)  break;
-        if (i >  3 && nidigits <  950 && poly_degree > 15)  break;
-        if (i >  4 && nidigits <  800 && poly_degree > 11)  break;
-        if (i >  8 && nidigits <  700 && poly_degree >  9)  break;
-        if (i > 16 && nidigits <  600 && poly_degree >  8)  break;
-      }
+      /* Make the continue-search vs. backtrack decision */
+      if (*pmaxH > 0 && poly_degree > *pmaxH)  break;
       mpz_set_si(mD, D);
       /* (D/N) must be 1, and we have to have a u,v solution */
       if (mpz_jacobi(mD, Ni) != 1)
@@ -790,23 +679,30 @@ static int ecpp_down(int i, mpz_t Ni, int facstage, IV* dlist, mpz_t* sfacs, int
 
       choose_m(mlist, D, u, v, Ni, t, t2);
       for (k = 0; k < 6; k++) {
-        facresult = check_for_factor2(q, mlist[k], minfactor, t, stage, sfacs, nsfacs, poly_degree);
+        int maxH = *pmaxH;
+        int minH = (nidigits <= 240) ? 6 : (nidigits+39)/40;
+        facresult = check_for_factor(q, mlist[k], minfactor, t, stage, sfacs, nsfacs, poly_degree);
         /* -1 = couldn't find, 0 = no big factors, 1 = found */
         if (facresult <= 0)
           continue;
         mpz_set(m, mlist[k]);
         if (verbose)
           { printf(" %d (%s %d)\n", D, (poly_type == 1) ? "Hilbert" : "Weber", poly_degree); fflush(stdout); }
+        if (maxH == 0) {
+          maxH = minH-1 + poly_degree;
+          if (facstage > 1)              /* We worked hard to get here, */
+            maxH = 2*maxH + 10;          /* try hard to make use of it. */
+        } else if (maxH > minH && maxH > (poly_degree+2)) {
+          maxH--;
+        }
         /* Great, now go down. */
-        downresult = ecpp_down(i+1, q, next_stage, dlist, sfacs, nsfacs, prooftextptr);
-        if (downresult == 0)     /* composite */
-          goto end_down;
+        downresult = ecpp_down(i+1, q, next_stage, &maxH, dlist, sfacs, nsfacs, prooftextptr);
+        /* Nothing found, look at more polys in the future */
+        if (downresult == 1 && *pmaxH > 0)  *pmaxH = maxH;
+
+        if (downresult == 0) goto end_down;   /* composite */
         if (downresult == 1) {   /* nothing found at this stage */
-          if (verbose) {
-            printf("%*sN[%d] (%d dig)", i, "", i, nidigits);
-            if (facstage > 1) printf(" FS %d", facstage);
-            fflush(stdout);
-          }
+          VERBOSE_PRINT_N(i, nidigits, *pmaxH, facstage);
           continue;
         }
 
@@ -821,7 +717,7 @@ static int ecpp_down(int i, mpz_t Ni, int facstage, IV* dlist, mpz_t* sfacs, int
            * messed up, and we've wasted a lot of time.  Sigh. */
           dlist[dnum] = -2; /* skip this D value from now on */
           if (verbose) gmp_printf("\n  Invalidated D = %d with N = %Zd\n", D, Ni);
-          downresult = 0;
+          downresult = 1;
           continue;
         }
         /* We found it was composite or proved it */
@@ -830,43 +726,28 @@ static int ecpp_down(int i, mpz_t Ni, int facstage, IV* dlist, mpz_t* sfacs, int
     } /* D */
   } /* fac stage */
   /* Nothing at this level */
-  downresult = 1;
-  if (verbose) { printf(" ---\n"); fflush(stdout); }
+  if (downresult != 1) croak("ECPP internal error: downresult is %d at end\n", downresult);
+  if (verbose) {
+    if (*pmaxH > 0) printf(" (max %d)", *pmaxH);
+    printf(" ---\n");
+    fflush(stdout);
+  }
+  if (*pmaxH > 0) *pmaxH = *pmaxH + 2;
 
 end_down:
 
   if (downresult == 2) {
     if (0 && verbose > 1) {
+      gmp_printf("\n");
       if (D == 1) {
-        gmp_printf("\n");
-        gmp_printf("Type BLS3\n");
-        gmp_printf("N  %Zd\n", Ni);
-        gmp_printf("Q  %Zd\n", q);
-        gmp_printf("A  %lu\n", (unsigned long) nm1a);
-        gmp_printf("\n");
-        fflush(stdout);
+        gmp_printf("Type BLS3\nN  %Zd\nQ  %Zd\nA  %"UVuf"\n", Ni, q, nm1a);
       } else if (D == -1) {
-        gmp_printf("\n");
-        gmp_printf("Type BLS15\n");
-        gmp_printf("N  %Zd\n", Ni);
-        gmp_printf("Q  %Zd\n", q);
-        gmp_printf("LP %ld\n", (signed long) np1lp);
-        gmp_printf("LQ %ld\n", (signed long) np1lq);
-        gmp_printf("\n");
-        fflush(stdout);
+        gmp_printf("Type BLS15\nN  %Zd\nQ  %Zd\nLP %"IVdf"\nLQ %"IVdf"\n", Ni, q, np1lp, np1lq);
       } else {
-        gmp_printf("\n");
-        gmp_printf("Type ECPP\n");
-        gmp_printf("N  %Zd\n", Ni);
-        gmp_printf("A  %Zd\n", a);
-        gmp_printf("B  %Zd\n", b);
-        gmp_printf("M  %Zd\n", m);
-        gmp_printf("Q  %Zd\n", q);
-        gmp_printf("X  %Zd\n", P.x);
-        gmp_printf("Y  %Zd\n", P.y);
-        gmp_printf("\n");
-        fflush(stdout);
+        gmp_printf("Type ECPP\nN  %Zd\nA  %Zd\nB  %Zd\nM  %Zd\nQ  %Zd\nX  %Zd\nY  %Zd\n", Ni, a, b, m, q, P.x, P.y);
       }
+      gmp_printf("\n");
+      fflush(stdout);
     }
     /* Prepend our proof to anything that exists. */
     if (prooftextptr != 0) {
@@ -901,6 +782,16 @@ end_down:
     }
   }
 
+  /* Ni passed BPSW, so it's highly unlikely to be composite */
+  if (downresult == 0) {
+    if (mpz_probab_prime_p(Ni, 2) == 0) {
+      gmp_printf("\n\n**** BPSW counter-example found?  ****\n**** N = %Zd ****\n\n", Ni);
+    } else {
+      /* Q was composite, but we don't seem to be. */
+      downresult = 1;
+    }
+  }
+
   mpz_clear(a);  mpz_clear(b);
   mpz_clear(u);  mpz_clear(v);
   mpz_clear(m);  mpz_clear(q);
@@ -910,7 +801,6 @@ end_down:
   for (k = 0; k < 6; k++)
     mpz_clear(mlist[k]);
 
-  /* TODO: proof text */
   return downresult;
 }
 
@@ -936,8 +826,9 @@ int _GMP_ecpp(mpz_t N, char** prooftextptr)
   nsfacs = 0;
   result = 1;
   for (fstage = 1; fstage < 20; fstage++) {
+    int maxH = 0;
     if (verbose && fstage == 3) gmp_printf("Working hard on: %Zd\n", N);
-    result = ecpp_down(0, N, fstage, dlist, sfacs, &nsfacs, prooftextptr);
+    result = ecpp_down(0, N, fstage, &maxH, dlist, sfacs, &nsfacs, prooftextptr);
     if (result != 1)
       break;
   }
@@ -950,297 +841,24 @@ int _GMP_ecpp(mpz_t N, char** prooftextptr)
 }
 
 
-
-
-/* This is the "factor and prove strategy" version, with no backtracking.
- * It tries to be efficient with factoring, but with a limited set of
- * discriminants and no backtracking it tends to get stuck factoring once
- * into the 250+ digit range.  More discriminants helps, but it just pushes
- * the problem out some.  The FAS version tends to be much faster.
- */
-typedef struct {
-  long D;
-  mpz_t m;
-  mpz_t q;
-} dmqlist_t;
-
-/* returns 2 if N is proven prime, 1 if probably prime, 0 if composite */
-int _GMP_ecpp_fps(mpz_t N, char** prooftextptr)
-{
-  mpz_t Ni, a, b, u, v, m, q, mD, t, t2, minfactor;
-  mpz_t mlist[6];
-  IV* dlist;
-  struct ec_affine_point P;
-  UV i, dnum;
-  long D, k, numQs, numDs, stage;
-  int result = 0;
-  int facresult, curveresult, niresult;
-  dmqlist_t *dmqlist;
-  char *proofstr, *proofptr;
-  size_t proofstr_size;
-  int verbose = get_verbose_level();
-
-  /* We must check gcd(N,6), let's check 2*3*5*7*11*13*17*19*23. */
-  if (mpz_gcd_ui(NULL, N, 223092870UL) != 1)
-    return _GMP_is_prob_prime(N);
-
-  init_ecpp_gcds();
-
-  if (prooftextptr) {
-    proofstr_size = 4096;
-    New(0, proofstr, proofstr_size, char);
-    *proofstr = 0;
-  } else {
-    proofstr_size = 0;
-    proofstr = 0;
-    proofptr = 0;
-  }
-  proofptr = proofstr;
-  mpz_init_set(Ni, N);
-  mpz_init(a); mpz_init(b); mpz_init(u); mpz_init(v);
-  mpz_init(m); mpz_init(q);
-  mpz_init(mD);
-  mpz_init(t); mpz_init(t2);
-  mpz_init(minfactor);
-  mpz_init(P.x); mpz_init(P.y);
-  dlist = poly_class_degrees(0);
-  for (k = 0; k < 6; k++)
-    mpz_init(mlist[k]);
-  /* TODO: make this 100 and realloc if needed */
-  New(0, dmqlist, 1000, dmqlist_t);
-
-  numQs = 0;
-  for (i = 0; i < 10000000; i++) {
-    if (verbose) {
-      int nidigits = mpz_sizeinbase(Ni, 10);
-      printf("  N[%d] (%d digits)\n", (int)i, nidigits);
-    }
-    /* Stop now if Ni is small enough to be known prime. */
-    k = _GMP_is_prob_prime(Ni);
-    if (k != 1) {
-      if (k == 2) result = 2;
-      goto end_ecpp;
-    }
-#ifdef USE_NM1
-    /* Try a simple n-1 test.  Note that Atkin and Morain discuss using
-     * n-1 and n+1 tests at each level, so this is fairly common.  It probably
-     * made even more sense in 1992 when the usual cutoff was a sieve.
-     *
-     * For not-big numbers, Effort 1 is nearly free, 2 is very fast, 3 is ok.
-     * The likelihood of this succeeding goes down rapidly with size, and the
-     * cost goes up.
-     * For effort 2, it is about 50% at 128 bits, 25% at 160, 10% at 190. */
-    {
-      UV log2ni = mpz_sizeinbase(Ni, 2);
-      int effort = (log2ni <= 128) ? 2
-                 : (log2ni <= 256) ? 1
-                 :                   0;
-      if (effort > 0) {
-        char* nm1_proof = 0;
-        /* -1 = composite, 0 = dunno, 1 = proved prime */
-        k = _GMP_primality_bls_nm1(Ni, effort, &nm1_proof);
-        if (k > 0) {
-          result = 2;
-          if (prooftextptr && nm1_proof != 0) {
-            size_t new_size = strlen(proofstr) + strlen(nm1_proof) + 1;
-            if ( new_size > proofstr_size ) {
-              proofstr_size = new_size;
-              Renew(proofstr, proofstr_size, char);
-            }
-            (void) strcat(proofstr, nm1_proof);
-          }
-        }
-        if (nm1_proof != 0)
-          Safefree(nm1_proof);
-        if (k != 0)
-          goto end_ecpp;
-      }
-    }
-#endif
-
-    /* Any factors q found must be strictly > minfactor.
-     * See Atkin and Morain, 1992, section 6.4 */
-    mpz_root(minfactor, Ni, 4);
-    mpz_add_ui(minfactor, minfactor, 1);
-    mpz_mul(minfactor, minfactor, minfactor);
-
-    niresult = 0;
-    numDs = 0;
-    if (numQs != 0) croak("assert numQs == 0\n");
-    for (stage = 1; stage <= 20; stage++) {
-      if (stage == 1) {
-        /* Walk the degree-sorted discriminants */
-        for (dnum = 0; dlist[dnum] > 0; dnum++) {
-          int poly_type;  /* just for debugging/verbose */
-          int poly_degree;
-          D = -dlist[dnum];
-          if ( (-D % 4) != 3 && (-D % 16) != 4 && (-D % 16) != 8 )
-            croak("Invalid discriminant '%ld' in list\n", D);
-          /* D must also be squarefree in odd divisors, but assume it. */
-          /* Make sure we can get a class polynomial for this D. */
-          poly_degree = poly_class_poly(D, NULL, &poly_type);
-          if (poly_degree == 0)  continue;
-          /* For debugging, one can put things here like:
-           * if poly_degree < 4) continue;
-           */
-          mpz_set_si(mD, D);
-          /* (D/N) must be 1, and we have to have a u,v solution */
-          if (mpz_jacobi(mD, Ni) != 1)
-            continue;
-          if ( ! modified_cornacchia(u, v, mD, Ni) )
-            continue;
-          numDs++;
-
-          if (verbose > 1)
-            { printf("  Candidate D %ld (%s poly)\n",
-                     D, (poly_type == 1) ? "Hilbert" : "Weber");
-              fflush(stdout); }
-
-          choose_m(mlist, D, u, v, Ni, t, t2);
-          for (k = 0; k < 6; k++) {
-            facresult = check_for_factor(q, mlist[k], minfactor, t, stage);
-            if (facresult == 0)
-              continue;
-            if (facresult > 0) {
-              mpz_set(m, mlist[k]);
-              if (verbose)
-                printf("  Found factor with D = %ld (class %d %s poly)\n",
-                       D, poly_degree, (poly_type == 1) ? "Hilbert" : "Weber");
-              curveresult = find_curve(a, b, P.x, P.y, D, m, q, Ni);
-              if (curveresult == 0)
-                goto end_ecpp;
-              /* If curveresult=1, then we're stumped.  Skip it? */
-              if (curveresult == 2) { niresult = 1; break; }
-            }
-            /* Add to list for future stages */
-            dmqlist[numQs].D = D;
-            mpz_init_set(dmqlist[numQs].m, mlist[k]);
-            mpz_init_set(dmqlist[numQs].q, q);
-            numQs++;
-          }
-          if (niresult) break;
-        }
-      } else {
-        /* We may want to consider sorting by size instead.  Certainly after
-         * stage 2 we're far more concerned with factoring than root finding. */
-        if (verbose)
-          { printf("  At stage %ld with %ld D values and %ld q values\n", stage, numDs, numQs); fflush(stdout); }
-        for (k = 0; k < numQs; k++) {
-          mpz_set(m, dmqlist[k].m);
-          facresult = check_for_factor(q, dmqlist[k].q, minfactor, t, stage);
-          if (facresult == 0) {
-            if (verbose > 1)
-              printf("  q has no large factor\n");
-            mpz_set(dmqlist[k].q, q);
-          } else if (facresult == -1) {
-            if (verbose > 1) {
-              unsigned long qdigits = mpz_sizeinbase(q, 10);
-              if (mpz_cmp(dmqlist[k].q,q) != 0)
-                printf("  q reduced (%lu digits)\n", qdigits);
-              else
-                printf("  q no luck (%lu digits)\n", qdigits);
-            }
-            mpz_set(dmqlist[k].q, q);
-          } else {
-            D = dmqlist[k].D;
-            curveresult = find_curve(a, b, P.x, P.y, D, m, q, Ni);
-            if (curveresult == 0)
-              goto end_ecpp;
-            /* If k=1, then we're stumped.  Skip it? */
-            if (curveresult == 2) { niresult = 1; break; }
-          }
-        }
-      }
-
-      if (niresult)  break;
-
-      if (stage == 1 && numDs == 0) {
-        /* If no Ds were found, make sure we don't have a composite. */
-        if (_GMP_miller_rabin_random(Ni, 10) == 0) {
-          if (verbose) gmp_printf("composite %Zd found in ecpp!\n", Ni);
-          goto end_ecpp;
-        }
-      }
-
-    } /* end stage loop */
-    for ( ; numQs > 0; numQs--) {
-      mpz_clear(dmqlist[numQs-1].m);
-      mpz_clear(dmqlist[numQs-1].q);
-    }
-    if (!niresult) {
-      if (verbose) printf("Giving up, sorry.\n");
-      result = 1;
-      goto end_ecpp;
-    }
-    if (verbose > 1) {
-      gmp_printf("N[%lu] = %Zd\n", (unsigned long) i, Ni);
-      gmp_printf("a = %Zd\n", a);
-      gmp_printf("b = %Zd\n", b);
-      gmp_printf("m = %Zd\n", m);
-      gmp_printf("q = %Zd\n", q);
-      gmp_printf("P = (%Zd, %Zd)\n", P.x, P.y);
-      fflush(stdout);
-    }
-    if (prooftextptr != 0) {
-      int prooflen = mpz_sizeinbase(Ni, 10) * 7 + 20;
-      size_t offset = proofptr - proofstr;
-      if ( (offset + 1 + prooflen) > proofstr_size ) {
-        proofstr_size += 4*prooflen;
-        Renew(proofstr, proofstr_size, char);
-        proofptr = proofstr + offset;
-      }
-      proofptr += gmp_sprintf(proofptr, "%Zd : ECPP : ", Ni);
-      proofptr += gmp_sprintf(proofptr,
-                              "%Zd %Zd %Zd %Zd (%Zd:%Zd)\n",
-                              a, b, m, q, P.x, P.y);
-    }
-    mpz_set(Ni, q);
-  }
-
-end_ecpp:
-  mpz_clear(Ni);
-  mpz_clear(a); mpz_clear(b); mpz_clear(u); mpz_clear(v);
-  mpz_clear(m); mpz_clear(q);
-  mpz_clear(mD);
-  mpz_clear(t); mpz_clear(t2);
-  mpz_clear(minfactor);
-  mpz_clear(P.x); mpz_clear(P.y);
-  Safefree(dlist);
-  for (k = 0; k < 6; k++)
-    mpz_clear(mlist[k]);
-  for ( ; numQs > 0; numQs--) {
-    mpz_clear(dmqlist[numQs-1].m);
-    mpz_clear(dmqlist[numQs-1].q);
-  }
-  Safefree(dmqlist);
-  /* Append our proof */
-  if (prooftextptr != 0) {
-    if (result != 2) {
-      Safefree(proofstr);
-    } else if (*prooftextptr == 0) {
-      *prooftextptr = proofstr;
-    } else {
-      Renew(*prooftextptr, strlen(*prooftextptr) + strlen(proofstr) + 1, char);
-      (void) strcat(*prooftextptr, proofstr);
-      Safefree(proofstr);
-    }
-  }
-
-  return result;
-}
-
 #ifdef STANDALONE_ECPP
 static void dieusage(char* prog) {
-  printf("ECPP-DJ version 1.0.  Dana Jacobsen\n\n");
+  printf("ECPP-DJ version 1.03.  Dana Jacobsen\n\n");
   printf("Usage: %s [options] <number>\n\n", prog);
   printf("Options:\n");
   printf("   -v     set verbose\n");
   printf("   -V     set extra verbose\n");
   printf("   -c     print certificate\n");
-  printf("   -nm1   use n-1 proof only\n");
+  printf("   -bpsw  use the extra strong BPSW test (probable prime test)\n");
+  printf("   -nm1   use n-1 proof only (BLS75 theorem 5)\n");
   printf("   -aks   use AKS for proof\n");
+#ifdef USE_APRCL
+  printf("   -aprcl use APR-CL for proof\n");
+#endif
   printf("   -help  this message\n");
-  exit(1);
+  printf("\n");
+  printf("Return codes: 0 prime, 1 composite, 2 prp, 3 error\n");
+  exit(3);
 }
 
 int main(int argc, char **argv)
@@ -1251,6 +869,7 @@ int main(int argc, char **argv)
   int do_aks = 0;
   int do_aprcl = 0;
   int do_bpsw = 0;
+  int retcode = 3;
   char* cert = 0;
 
   if (argc < 2) dieusage(argv[0]);
@@ -1320,10 +939,12 @@ int main(int argc, char **argv)
     /* printf("(%d digit) ", (int)mpz_sizeinbase(n, 10)); */
     if (isprime == 0) {
       printf("COMPOSITE\n");
+      retcode = 1;
     } else if (isprime == 1) {
-      /* We really shouldn't ever see this. */
+      /* This would normally only be from BPSW */
       printf("PROBABLY PRIME\n");
-    } else {
+      retcode = 2;
+    } else if (isprime == 2) {
       if (do_printcert) {
         gmp_printf("[MPU - Primality Certificate]\n");
         gmp_printf("Version 1.0\n");
@@ -1335,6 +956,10 @@ int main(int argc, char **argv)
       } else {
         printf("PRIME\n");
       }
+      retcode = 0;
+    } else {
+      /* E.g. APRCL returns -1 for error */
+      croak("Unknown return code, probable error.\n");
     }
     if (cert != 0) {
       Safefree(cert);
@@ -1343,6 +968,6 @@ int main(int argc, char **argv)
   }
   mpz_clear(n);
   _GMP_destroy();
-  return 0;
+  return retcode;
 }
 #endif
