@@ -11,7 +11,7 @@ extern int  _GMP_miller_rabin(mpz_t n, mpz_t a);
 extern int  _GMP_is_lucas_pseudoprime(mpz_t n, int strength);
 extern int  _GMP_is_almost_extra_strong_lucas_pseudoprime(mpz_t n, UV incr);
 extern int  _GMP_is_frobenius_underwood_pseudoprime(mpz_t n);
-extern int  _GMP_miller_rabin_random(mpz_t n, UV numbases);
+extern int  _GMP_miller_rabin_random(mpz_t n, UV numbases, char* seedstr);
 
 extern void _GMP_lucas_seq(mpz_t U, mpz_t V, mpz_t n, IV P, IV Q, mpz_t k,
                            mpz_t Qk, mpz_t t);
